package main

import (
	"crypto/rand"
	"encoding/binary"
	"flag"
	"fmt"
	"io"
	"log"
	"os"

	"github.com/netfoundry/secretstream"
	"golang.org/x/crypto/chacha20poly1305"
	"golang.org/x/crypto/argon2"
	"golang.org/x/term"
)

const (
	cream_magic               = "CREAM\x01"
	cream_version             = 0x10

	// internal stream cipher buffer
	xchacha20poly1305_buf_len = 4096

	// recommended for memory constrained environments, as per RFC9106
	argon2id_memory_cost      = 65536 // 64 Kib
	argon2id_time_cost        = 3
	argon2id_threads          = 4
	argon2id_salt_len         = 16
)

type Header struct {
	Cream_magic [6]byte
	Cream_version uint16
	Xchacha20poly1305_buf_len uint32
	Argon2id_memory_cost uint32
	Argon2id_time_cost uint32
	Argon2id_threads uint32
	Argon2id_salt [16]byte
}

func usage() {
	fmt.Printf("usage: %s [-hed] [-b size] [-j thread] [-t time] [-m memory] [-s salt] [-f file]\n", os.Args[0])
	os.Exit(2)
}

func readheader(f *os.File, h *Header) {
	err := binary.Read(f, binary.BigEndian, h)
	if err != nil {
		log.Fatal(err)
	}

	if string(h.Cream_magic[:]) != string(cream_magic) {
		log.Fatal("Invalid format", h.Cream_magic)
	}

	if h.Cream_version != cream_version {
		log.Fatal("Version %d not supported", h.Cream_version)
	}
}

func deriv(pw []byte, key *[]byte, time, mem, job uint32, salt []byte) {
	klen := len(*key)
	*key = argon2.IDKey(pw, salt, time, mem, uint8(job), uint32(klen))
}

func encrypt(in *os.File, out *os.File, key []byte, h Header) {
	var tag byte
	buf := make([]byte, h.Xchacha20poly1305_buf_len)

	enc, nonce, err := secretstream.NewEncryptor(key)
	if err != nil {
		log.Fatal(err)
	}

	binary.Write(out, binary.BigEndian, h)
	if err != nil {
		log.Fatal(err)
	}

	out.Write(nonce)
	if err != nil {
		log.Fatal(err)
	}

	loop := 1
	tag = secretstream.TagMessage
	for loop > 0 {
		n, err := in.Read(buf)
		if err == io.EOF || n < len(buf) {
			tag = secretstream.TagFinal
			loop = 0
		} else if err != nil {
			log.Fatal(err)
		}
		cipher, err := enc.Push(buf[:n], tag)
		if err != nil {
			log.Fatal(err)
		}

		out.Write(cipher)
		if err != nil {
			log.Fatal(err)
		}
	}
}

func decrypt(in *os.File, out *os.File, skip int, key []byte, blen uint32) {
	buf := make([]byte, blen+secretstream.StreamABytes)
	nonce := make([]byte, secretstream.StreamHeaderBytes)

	// Skip beginning of file which (supposedly) contains the cream header
	in.Seek(int64(skip), os.SEEK_SET)
	_, err := in.Read(nonce)
	if err != nil {
		log.Fatal(err)
	}

	dec, err := secretstream.NewDecryptor(key, nonce)
	if err != nil {
		log.Fatal(err)
	}

	loop := 1
	for loop > 0 {
		n, err := in.Read(buf)
		if err != nil && err != io.EOF {
			log.Fatal(err)
		}

		plain, tag, err := dec.Pull(buf[:n])
		if err != nil {
			log.Fatal(err)
		}

		if tag == secretstream.TagFinal {
			loop = 0
		}
		out.Write(plain)
	}
}

func main() {
	var err error
	var key, pass []byte
	var filename, saltfile string
	var dflag, eflag bool

        var time uint64
        var memory uint64
        var jobnum uint64
        var blksiz uint64
	var header Header

	in := os.Stdin
	out := os.Stdout

	copy(header.Cream_magic[:], cream_magic)
	header.Cream_version = cream_version

	key = make([]byte, chacha20poly1305.KeySize)

	flag.StringVar(&filename, "f", "", "Encrypt/decrypt to/from file name")
	flag.StringVar(&saltfile, "s", "", "Read salt from file (encrypt-only)")
	flag.BoolVar(&eflag, "e", false, "encrypt input (default)")
	flag.BoolVar(&dflag, "d", false, "decrypt input")

	// xchacha20/argon2id parameters
	flag.Uint64Var(&memory, "m", argon2id_memory_cost, "Memory cost")
	flag.Uint64Var(&time, "t", argon2id_time_cost, "Time cost")
	flag.Uint64Var(&jobnum, "j", argon2id_threads, "Parallel threads")
	flag.Uint64Var(&blksiz, "b", xchacha20poly1305_buf_len, "Buffer size")

	flag.Usage = usage
	flag.Parse()

	header.Argon2id_memory_cost = uint32(memory)
	header.Argon2id_time_cost = uint32(time)
	header.Argon2id_threads = uint32(jobnum)
	header.Xchacha20poly1305_buf_len = uint32(blksiz)

	if eflag && dflag {
		log.Fatal("Cannot use encryption and decryption at the same time")
	}

	args := flag.Args()
	if len(args) > 0 {
		usage()
	}

	// Prompt user for password
	tty, err := os.OpenFile("/dev/tty", os.O_RDWR, 0644)
	if err != nil {
		log.Fatal(err)
	}
	defer tty.Close()
	fmt.Fprint(tty, "password:")
	pass, _ = term.ReadPassword(int(tty.Fd()))
	fmt.Fprintln(tty, "")

	if dflag {
		if len(filename) > 0 {
			in, err = os.Open(filename)
			if err != nil {
				log.Fatal(err)
			}
			defer in.Close()
		}

		readheader(in, &header)
		deriv(pass, &key,
			header.Argon2id_time_cost,
			header.Argon2id_memory_cost,
			header.Argon2id_threads,
			header.Argon2id_salt[:])
		decrypt(in, out, binary.Size(header), key, header.Xchacha20poly1305_buf_len)
	} else {
		if len(saltfile) > 0 {
			// Read salt from any manually specified file…
			f, err := os.Open(saltfile)
			if err != nil {
				log.Fatal(err)
			}
			_, err = f.Read(header.Argon2id_salt[:])
			f.Close()
		} else {
			// … or generate a random one
			_, err = rand.Read(header.Argon2id_salt[:])
			if err != nil {
				log.Fatal(err)
			}
		}
		deriv(pass, &key,
			header.Argon2id_time_cost,
			header.Argon2id_memory_cost,
			header.Argon2id_threads,
			header.Argon2id_salt[:])
		if len(filename) > 0 {
			out, err := os.OpenFile(filename, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0600)
			if err != nil {
				log.Fatal(err)
			}
			defer out.Close()
		}
		encrypt(in, out, key, header)
	}
}
